package java.lang.reflect;

import java.util.Iterator;
import java.lang.ref.WeakReference;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.BiFunction;
import java.util.function.Supplier;
import java.util.concurrent.ConcurrentMap;
import java.lang.ref.ReferenceQueue;

final class WeakCache<K, P, V>
{
    private final ReferenceQueue<K> refQueue;
    private final ConcurrentMap<Object, ConcurrentMap<Object, Supplier<V>>> map;
    private final ConcurrentMap<Supplier<V>, Boolean> reverseMap;
    private final BiFunction<K, P, ?> subKeyFactory;
    private final BiFunction<K, P, V> valueFactory;
    
    public WeakCache(BiFunction<K, P, ?> obj, BiFunction<K, P, V> obj2) {
        this.refQueue = new ReferenceQueue();
        this.map = new ConcurrentHashMap();
        this.reverseMap = new ConcurrentHashMap();
        this.subKeyFactory = (BiFunction<K, P, ?>)Objects.requireNonNull(obj);
        this.valueFactory = (BiFunction<K, P, V>)Objects.requireNonNull(obj2);
    }
    
    public V get(K k, P obj) {
        Objects.requireNonNull(obj);
        this.expungeStaleEntries();
        Object value = WeakCache.CacheKey.valueOf(k, this.refQueue);
        ConcurrentMap concurrentMap = (ConcurrentMap)this.map.get(value);
        if (concurrentMap == null) {
            ConcurrentMap<Object, Object> concurrentMap2 = (ConcurrentMap)this.map.putIfAbsent(value, concurrentMap = new ConcurrentHashMap());
            if (concurrentMap2 != null) {
                concurrentMap = concurrentMap2;
            }
        }
        Object requireNonNull = Objects.requireNonNull(this.subKeyFactory.apply(k, obj));
        Supplier<Object> supplier = (Object)concurrentMap.get(requireNonNull);
        Object o = null;
        Object value2;
        while (true) {
            if (supplier != null) {
                value2 = supplier.get();
                if (value2 != null) {
                    break;
                }
            }
            if (o == null) {
                o = new WeakCache.Factory(k, obj, requireNonNull, concurrentMap);
            }
            if (supplier == null) {
                supplier = (Supplier)concurrentMap.putIfAbsent(requireNonNull, o);
                if (supplier != null) {
                    continue;
                }
                supplier = o;
            }
            else if (concurrentMap.replace(requireNonNull, supplier, o)) {
                supplier = o;
            }
            else {
                supplier = (Supplier)concurrentMap.get(requireNonNull);
            }
        }
        return value2;
    }
    
    public boolean containsValue(V obj) {
        Objects.requireNonNull(obj);
        this.expungeStaleEntries();
        return this.reverseMap.containsKey(new WeakCache.LookupValue(obj));
    }
    
    public int size() {
        this.expungeStaleEntries();
        return this.reverseMap.size();
    }
    
    private void expungeStaleEntries() {
        WeakCache.CacheKey<Object> cacheKey;
        while ((cacheKey = (WeakCache.CacheKey)this.refQueue.poll()) != null) {
            cacheKey.expungeFrom(this.map, this.reverseMap);
        }
    }
    
    private final class Factory implements Supplier<V>
    {
        private final K key;
        private final P parameter;
        private final Object subKey;
        private final ConcurrentMap<Object, Supplier<V>> valuesMap;
        
        Factory(K key, P parameter, Object subKey, ConcurrentMap<Object, Supplier<V>> valuesMap) {
            this.key = key;
            this.parameter = parameter;
            this.subKey = subKey;
            this.valuesMap = valuesMap;
        }
        
        @Override
        public synchronized V get() {
            if ((Supplier)this.valuesMap.get(this.subKey) != this) {
                return null;
            }
            Object requireNonNull = null;
            try {
                requireNonNull = Objects.requireNonNull(WeakCache.this.valueFactory.apply(this.key, this.parameter));
            }
            finally {
                if (requireNonNull == null) {
                    this.valuesMap.remove(this.subKey, this);
                }
            }
            assert requireNonNull != null;
            WeakCache.CacheValue<V> cacheValue = new CacheValue(requireNonNull);
            WeakCache.this.reverseMap.put(cacheValue, Boolean.TRUE);
            if (!this.valuesMap.replace(this.subKey, this, cacheValue)) {
                throw new AssertionError("Should not reach here");
            }
            return requireNonNull;
        }
    }
    
    private static final class LookupValue<V> implements WeakCache.Value<V>
    {
        private final V value;
        
        LookupValue(V value) {
            this.value = value;
        }
        
        @Override
        public V get() {
            return this.value;
        }
        
        @Override
        public int hashCode() {
            return System.identityHashCode(this.value);
        }
        
        @Override
        public boolean equals(Object o) {
            return o == this || (o instanceof WeakCache.Value && this.value == ((WeakCache.Value)o).get());
        }
    }
    
    private static final class CacheValue<V> extends WeakReference<V> implements WeakCache.Value<V>
    {
        private final int hash;
        
        CacheValue(V referent) {
            super(referent);
            this.hash = System.identityHashCode(referent);
        }
        
        @Override
        public int hashCode() {
            return this.hash;
        }
        
        @Override
        public boolean equals(Object o) {
            Object value;
            return o == this || (o instanceof WeakCache.Value && (value = this.get()) != null && value == ((WeakCache.Value)o).get());
        }
    }
    
    private static final class CacheKey<K> extends WeakReference<K>
    {
        private static final Object NULL_KEY;
        private final int hash;
        
        static <K> Object valueOf(K k, ReferenceQueue<K> referenceQueue) {
            return (k == null) ? WeakCache.CacheKey.NULL_KEY : new WeakCache.CacheKey(k, referenceQueue);
        }
        
        private CacheKey(K referent, ReferenceQueue<K> q) {
            super(referent, q);
            this.hash = System.identityHashCode(referent);
        }
        
        @Override
        public int hashCode() {
            return this.hash;
        }
        
        @Override
        public boolean equals(Object o) {
            Object value;
            return o == this || (o != null && o.getClass() == this.getClass() && (value = this.get()) != null && value == ((WeakCache.CacheKey)o).get());
        }
        
        void expungeFrom(ConcurrentMap<?, ? extends ConcurrentMap<?, ?>> concurrentMap, ConcurrentMap<?, Boolean> concurrentMap2) {
            ConcurrentMap<Object, Object> concurrentMap3 = (ConcurrentMap)concurrentMap.remove(this);
            if (concurrentMap3 != null) {
                Iterator<Object> iterator = concurrentMap3.values().iterator();
                while (iterator.hasNext()) {
                    concurrentMap2.remove(iterator.next());
                }
            }
        }
        
        static {
            NULL_KEY = new Object();
        }
    }
    
    private interface Value<V> extends Supplier<V>
    {
    }
}
